@@ -252,7 +252,7 @@ ne_sock_connect (ne_socket *sock, const ne_inet_addr *addr,
 		return NE_SOCK_ERROR;
 	  
 	sock->socket = gnome_vfs_inet_connection_to_socket (sock->connection);	
-	sock->socket_buffer = gnome_vfs_inet_connection_to_socket_buffer (sock->connection);
+	sock->socket_buffer = gnome_vfs_socket_buffer_new (sock->socket);
 			 
 	return 0;
 }
@@ -426,8 +426,6 @@ ne_sock_close (ne_socket *sock)
 
 	peek_cancellation (cancellation);
 
-	gnome_vfs_socket_free (sock->socket);	   
-
 	if (sock->socket_buffer) {
 		gnome_vfs_socket_buffer_flush (sock->socket_buffer,
 					       cancellation);
