@@ -74,14 +74,11 @@ typedef struct {
 } SmbServerCacheEntry;
 
 typedef struct {
-	char *server_name; /* Server or workgroup */
-	char *share_name; /* NULL for just the server or a workgroup,
-			     NULL used as default for shares if there is no
-			     specific one for the share */
-	
 	char *username;
 	char *domain;
-} SmbDefaultUser;
+        char *password;
+        time_t stamp;
+} SmbCachedUser;
 
 static GMutex *smb_lock;
 
@@ -89,11 +86,13 @@ static SMBCCTX *smb_context = NULL;
 
 static GHashTable *server_cache = NULL;
 
+static GHashTable *user_cache = NULL;
+
 #define SMB_BLOCK_SIZE (32*1024)
 
-/* Reap unused server connections after 30 minutes */
-#define SERVER_CACHE_REAP_TIMEOUT (30*60)
-static guint server_cache_reap_timeout = 0;
+/* Reap unused server connections and user cache after 30 minutes */
+#define CACHE_REAP_TIMEOUT (30 * 60)
+static guint cache_reap_timeout = 0;
 
 /* The magic "default workgroup" hostname */
 #define DEFAULT_WORKGROUP_NAME "X-GNOME-DEFAULT-WORKGROUP"
@@ -105,8 +104,6 @@ static guint server_cache_reap_timeout = 0;
 static GHashTable *workgroups = NULL;
 static time_t workgroups_timestamp = 0;
 
-static GHashTable *default_user_hashtable = NULL;
-
 /* Authentication ----------------------------------------------------------- */
 
 #define SMB_AUTH_STATE_PREFILLED	0x00000010 	/* Have asked gnome-auth for prefilled auth */
@@ -133,6 +130,9 @@ typedef struct _SmbAuthContext {
 	gchar *use_user;		/* Set by perform_authentication */
 	gchar *use_domain;		/* Set by perform_authentication */
 	gchar *use_password;		/* Set by perform_authentication */
+
+	gboolean cache_added;           /* Set when cache is added to during authentication */
+	gboolean cache_used;            /* Set when cache is used during authentication */
 	
 } SmbAuthContext;
 
@@ -141,9 +141,6 @@ static int  perform_authentication (SmbAuthContext *actx);
 
 static SmbAuthContext *current_auth_context = NULL;
 
-/* Used to detect failed logins */
-static gboolean cache_access_failed = FALSE;
-
 static void auth_callback (const char *server_name, const char *share_name,
 		     	   char *domain, int domainmaxlen,
 		     	   char *username, int unmaxlen,
@@ -194,6 +191,15 @@ string_compare (const char *a, const char *b)
 	}
 }
 
+static gchar*
+string_realloc (gchar *dest, const gchar *src)
+{
+        if (string_compare (src, dest))
+                return dest;
+        g_free (dest);
+        return string_dup_nzero (src);
+}
+
 static gboolean
 server_equal (gconstpointer  v1,
 	      gconstpointer  v2)
@@ -246,46 +252,13 @@ server_free (SmbServerCacheEntry *entry)
 	g_free (entry);
 }
 
-static gboolean
-default_user_equal (gconstpointer  v1,
-		    gconstpointer  v2)
-{
-	const SmbDefaultUser *e1, *e2;
-
-	e1 = v1;
-	e2 = v2;
-	
-	return (string_compare (e1->server_name, e2->server_name) &&
-		string_compare (e1->share_name, e2->share_name));
-}
-
-static guint
-default_user_hash (gconstpointer  v)
-{
-	const SmbDefaultUser *e;
-	guint hash;
-
-	e = v;
-	hash = 0;
-	if (e->server_name) {
-		hash = g_str_hash (e->server_name);
-	}
-	if (e->share_name) {
-		hash = hash ^ g_str_hash (e->share_name);
-	}
-	return hash;
-}
-
 static void
-default_user_free (SmbDefaultUser *entry)
+user_free (SmbCachedUser *user)
 {
-	g_free (entry->server_name);
-	g_free (entry->share_name);
-	
-	g_free (entry->username);
-	g_free (entry->domain);
-
-	g_free (entry);
+        g_free (user->username);
+        g_free (user->domain);
+        g_free (user->password);
+        g_free (user);
 }
 
 static void
@@ -301,19 +274,34 @@ add_old_servers (gpointer key,
 
 	entry = key;
 	array = user_data;
-	if (now > entry->last_time + SERVER_CACHE_REAP_TIMEOUT ||
-	    now < entry->last_time) {
+	if (now > entry->last_time + CACHE_REAP_TIMEOUT || now < entry->last_time)
 		g_ptr_array_add (array, entry->server);
-	}
 }
 
+static gboolean
+reap_user (gpointer key, gpointer value, gpointer user_data)
+{
+        SmbCachedUser *user = (SmbCachedUser*)value;
+        time_t now =  time (NULL);
+
+        if (now > user->stamp + CACHE_REAP_TIMEOUT || now < user->stamp) 
+                return TRUE;
+        return FALSE;              
+}
 
 static gboolean
-server_cache_reap_cb (void)
+cache_reap_cb (void)
 {
-     	int size;
 	GPtrArray *servers;
+        gboolean ret;
+        int size;
 	int i;
+        
+        /* Don't deadlock here, this is callback and we're not completely
+         * sure when we'll be called */
+        if (!g_mutex_trylock (smb_lock))
+                return TRUE;
+        DEBUG_SMB(("LOCK %s\n", G_GNUC_PRETTY_FUNCTION));
 
 	size = g_hash_table_size (server_cache);
 	servers = g_ptr_array_sized_new (size);
@@ -327,21 +315,26 @@ server_cache_reap_cb (void)
 	}
 
 	g_ptr_array_free (servers, TRUE);
+     
+        /* Cleanup users */
+        g_hash_table_foreach_remove (user_cache, reap_user, NULL);
 
-	if (g_hash_table_size (server_cache) == 0) {
-		server_cache_reap_timeout = 0;
-		return FALSE;
-	} else {
-		return TRUE;
-	}
+        ret = !(g_hash_table_size (server_cache) == 0 && 
+                g_hash_table_size (user_cache) == 0);
+        if (!ret)
+                cache_reap_timeout = 0;
+
+        UNLOCK_SMB();
+
+        return ret;        
 }
 
 static void
-schedule_server_cache_reap (void)
+schedule_cache_reap (void)
 {
-	if (server_cache_reap_timeout == 0) {
-		server_cache_reap_timeout = g_timeout_add (SERVER_CACHE_REAP_TIMEOUT*1000,
-							   (GSourceFunc)server_cache_reap_cb, NULL);
+	if (cache_reap_timeout == 0) {
+		cache_reap_timeout = g_timeout_add (CACHE_REAP_TIMEOUT * 1000,
+						    (GSourceFunc)cache_reap_cb, NULL);
 	}
 }
 
@@ -351,13 +344,11 @@ add_cached_server (SMBCCTX *context, SMBCSRV *new,
 		   const char *domain, const char *username)
 {
 	SmbServerCacheEntry *entry = NULL;
-	GnomeVFSToplevelURI *toplevel;
-	SmbDefaultUser *default_user;
 
-	DEBUG_SMB(("add_cached_server: server: %s, share: %s, domain: %s, user: %s\n",
+	DEBUG_SMB(("[auth] adding cached server: server: %s, share: %s, domain: %s, user: %s\n",
 		   server_name, share_name, domain, username));
 	
-	schedule_server_cache_reap ();
+	schedule_cache_reap ();
 	
 	entry = g_new0 (SmbServerCacheEntry, 1);
 	
@@ -370,23 +361,7 @@ add_cached_server (SMBCCTX *context, SMBCSRV *new,
 	entry->last_time = time (NULL);
 
 	g_hash_table_insert (server_cache, entry, entry);
-
-	cache_access_failed = FALSE;
-
-	if (current_auth_context && current_auth_context->uri != NULL) {
-		toplevel = (GnomeVFSToplevelURI *)current_auth_context->uri;
-
-		if (toplevel->user_name == NULL ||
-		    toplevel->user_name[0] == 0) {
-			default_user = g_new0 (SmbDefaultUser, 1);
-			default_user->server_name = string_dup_nzero (server_name);
-			default_user->share_name = string_dup_nzero (share_name);
-			default_user->username = string_dup_nzero (username);
-			default_user->domain = string_dup_nzero (domain);
-			g_hash_table_replace (default_user_hashtable, default_user, default_user);
-		}
-	}
-
+	current_auth_context->cache_added = TRUE;
 	return 0;
 }
 
@@ -406,11 +381,12 @@ get_cached_server (SMBCCTX * context,
 	res = g_hash_table_lookup (server_cache, &entry);
 
 	if (res != NULL) {
-		cache_access_failed = FALSE;
+		DEBUG_SMB(("got cached server: server: %s, share: %s, domain: %s, user: %s\n",
+		   	  server_name, share_name, domain, username));
+		current_auth_context->cache_used = TRUE;
 		res->last_time = time (NULL);
 		return res->server;
 	}
-	cache_access_failed = TRUE;
 	return NULL;
 }
 
@@ -420,20 +396,12 @@ remove_server  (gpointer key,
 		gpointer user_data)
 {
 	SmbServerCacheEntry *entry;
-	SmbDefaultUser default_user;
 	SMBCSRV *server;
 	
 	entry = key;
 	server = user_data;
 
 	if (entry->server == user_data) {
-		default_user.server_name = entry->server_name;
-		default_user.share_name = entry->share_name;
-		default_user.username = entry->username;
-		default_user.domain = entry->domain;
-		
-		g_hash_table_remove (default_user_hashtable, &default_user);
-		
 		entry->server = NULL;
 		return TRUE;
 	} 
@@ -654,23 +622,18 @@ try_init (void)
 #if defined(SMB_CTX_FLAG_USE_KERBEROS) && defined(SMB_CTX_FLAG_FALLBACK_AFTER_KERBEROS)
 		smb_context->flags |= SMB_CTX_FLAG_USE_KERBEROS | SMB_CTX_FLAG_FALLBACK_AFTER_KERBEROS;
 #endif
-#if defined(SMBCCTX_FLAG_NO_AUTO_ANON)
-		smb_context->flags |= SMBCCTX_FLAG_NO_AUTO_ANON;
+#if defined(SMBCCTX_FLAG_NO_AUTO_ANONYMOUS_LOGON)
+		smb_context->flags |= SMBCCTX_FLAG_NO_AUTO_ANONYMOUS_LOGON;
 #endif
 #endif		
 	}
 
-	server_cache = g_hash_table_new_full (server_hash,
-					      server_equal,
-					      (GDestroyNotify)server_free,
-					      NULL);
-	workgroups = g_hash_table_new_full (g_str_hash,
-					    g_str_equal,
+	server_cache = g_hash_table_new_full (server_hash, server_equal,
+					      (GDestroyNotify)server_free, NULL);
+	workgroups = g_hash_table_new_full (g_str_hash, g_str_equal,
 					    g_free, NULL);
-	
-	default_user_hashtable = g_hash_table_new_full (default_user_hash,
-							default_user_equal,
-							(GDestroyNotify)default_user_free, NULL);
+	user_cache = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                            g_free, (GDestroyNotify)user_free);
 	
 	UNLOCK_SMB();
 
@@ -685,19 +648,83 @@ try_init (void)
 
 /* Authentication ----------------------------------------------------------- */
 
-static char*
-get_auth_display_uri (SmbAuthContext *actx)
+static const gchar*
+get_auth_display_share (SmbAuthContext *actx)
 {
-        if (actx->uri != NULL)
-		return gnome_vfs_uri_to_string (actx->uri, 0);
-    	else
-        	return g_strdup_printf ("smb://%s%s%s%s", 
-					actx->for_server ? actx->for_server : "", 
-					actx->for_server ? "/" : "",
-					actx->for_share ? actx->for_share : "",
-					actx->for_share ? "/" : "");
+        return string_compare (actx->for_share, "IPC$") ? 
+                        NULL : actx->for_share;
 }
 
+static char*
+get_auth_display_uri (SmbAuthContext *actx, gboolean machine)
+{
+        if (actx->uri != NULL && !machine)
+                return gnome_vfs_uri_to_string (actx->uri, 0);
+        else {
+                const gchar *share = get_auth_display_share (actx);
+                return g_strdup_printf ("smb://%s%s%s%s", 
+                                        actx->for_server ? actx->for_server : "", 
+                                        actx->for_server ? "/" : "",
+                                        share && !machine ? share : "",
+                                        share && !machine ? "/" : "");
+        }
+}
+
+static void
+update_user_cache (SmbAuthContext *actx, gboolean with_share)
+{
+        SmbCachedUser *user;
+        gchar *key;
+        
+        g_return_if_fail (actx->for_server != NULL);
+        
+        key = g_strdup_printf ("%s/%s", actx->for_server, with_share ? actx->for_share : "");
+        user = (SmbCachedUser*)g_hash_table_lookup (user_cache, key);
+        
+        DEBUG_SMB(("[auth] Saved in cache: %s = %s:%s@%s\n", key, actx->use_user, actx->use_domain, actx->use_password));
+        
+        if (!user) {
+                user = g_new0 (SmbCachedUser, 1);
+                g_hash_table_replace (user_cache, key, user);
+                schedule_cache_reap ();
+        } else {
+                g_free (key);
+        }
+
+        user->domain = string_realloc (user->domain, actx->use_domain);
+        user->username = string_realloc (user->username, actx->use_user);
+        user->password = string_realloc (user->password, actx->use_password);
+        user->stamp = time (NULL);
+}
+
+static gboolean
+lookup_user_cache (SmbAuthContext *actx, gboolean with_share, gboolean with_user)
+{
+        SmbCachedUser *user;
+        gchar *key;
+       
+        g_return_val_if_fail (actx->for_server != NULL, FALSE);
+       
+        key = g_strdup_printf ("%s/%s", actx->for_server, with_share ? actx->for_share : "");
+        user = (SmbCachedUser*)g_hash_table_lookup (user_cache, key);
+        g_free (key);
+       
+        if (user) {
+                /* If the caller specified we already have a user name double check that... */
+                if (with_user && (!string_compare (user->username, actx->use_user) ||
+                                  !string_compare (user->domain, actx->use_domain)))
+                    return FALSE;
+
+                actx->use_user = string_realloc (actx->use_user, user->username);
+                actx->use_domain = string_realloc (actx->use_domain, user->domain);
+                actx->use_password = string_realloc (actx->use_password, user->password);
+                DEBUG_SMB(("[auth] Looked up in cache: %s:%s@%s\n", actx->use_user, actx->use_domain, actx->use_password));
+                return TRUE;
+        }
+        
+        return FALSE;
+}      
+
 static gboolean
 initial_authentication (SmbAuthContext *actx)
 {
@@ -706,8 +733,8 @@ initial_authentication (SmbAuthContext *actx)
 	GnomeVFSToplevelURI *toplevel_uri;
 	SmbServerCacheEntry server_lookup;
 	SmbServerCacheEntry *server;
-	SmbDefaultUser *default_user;
-	SmbDefaultUser lookup;
+        gboolean found_user = FALSE;
+        gboolean preset_user = FALSE;
 	char *tmp;
 
 	DEBUG_SMB(("[auth] Initial authentication lookups\n"));
@@ -733,44 +760,41 @@ initial_authentication (SmbAuthContext *actx)
 			actx->use_domain = NULL;
 			DEBUG_SMB(("[auth] User from URI: %s\n", actx->use_user));
 		}
+                preset_user = TRUE;
+	} 
 
-	/* Lookup a default user and domain */
-	} else {
-		
-		/* lookup default user/domain */
-		lookup.server_name = actx->for_server;
-		lookup.share_name = actx->for_share;
+        if (lookup_user_cache (actx, TRUE, preset_user) ||
+            lookup_user_cache (actx, FALSE, preset_user))
+                found_user = TRUE;
 
-		default_user = g_hash_table_lookup (default_user_hashtable, &lookup);
-		if (default_user != NULL) {
-			g_free (actx->use_user);
-			actx->use_user = string_dup_nzero (default_user->username);
-			g_free (actx->use_domain);
-			actx->use_domain = string_dup_nzero (default_user->domain);
-			DEBUG_SMB(("[auth] Looked up default: %s@%s\n", actx->use_user, actx->use_domain));
-		}
-	}
-	
-	/* Lookup the password in our internal cache */
-	if (actx->use_user != NULL && actx->use_user[0] != 0) {
-		
-		server_lookup.server_name = (char*)actx->for_server;
-		server_lookup.share_name = (char*)actx->for_share;
-		server_lookup.username = (char*)actx->use_user;
-		server_lookup.domain = (char*)actx->use_domain;
-		
-		server = g_hash_table_lookup (server_cache, &server_lookup);
-		if (server != NULL) {
-			/* Server is in cache already, no need to get password */
-			g_free (actx->use_password);
-			actx->use_password = g_strdup ("");
-			DEBUG_SMB(("[auth] Using login info for '%s@%s' from our internal cache\n", actx->use_user, actx->use_domain));
-			return TRUE;
-		}
-	}	
+        if (found_user || preset_user) {
+        	/* Lookup the server in our internal cache */
+        	server_lookup.server_name = (char*)actx->for_server;
+        	server_lookup.share_name = (char*)actx->for_share;
+        	server_lookup.username = (char*)actx->use_user;
+        	server_lookup.domain = (char*)actx->use_domain;
+        		
+        	server = g_hash_table_lookup (server_cache, &server_lookup);
+        	if (server == NULL) {
+                 
+                        /* If a blank user, try 'guest' */
+                        if (!actx->use_user) {
+                                server_lookup.username = "guest";
+                                server_lookup.domain = NULL;
+                                server = g_hash_table_lookup (server_cache, &server_lookup);
+                        }
+                }
+                
+                /* Server is in cache already, no need to get password */
+                if (server != NULL) {
+        		g_free (actx->use_password);
+        		actx->use_password = g_strdup ("");
+        		DEBUG_SMB(("[auth] Using connection for '%s@%s' from our server cache\n", actx->use_user, actx->use_domain));
+                        found_user = TRUE;
+        	}
+        }
 
-	/* TODO: Complete full cache authentication loading */
-	return FALSE;
+	return found_user;
 }
 
 static gboolean
@@ -786,10 +810,10 @@ prefill_authentication (SmbAuthContext *actx)
 	g_return_val_if_fail (actx->for_server != NULL, FALSE);	
 
 	memset (&in_args, 0, sizeof (in_args));
-	in_args.uri = get_auth_display_uri (actx);
+	in_args.uri = get_auth_display_uri (actx, FALSE);
 	in_args.protocol = "smb";
 	in_args.server = (char*)actx->for_server;
-	in_args.object = (char*)actx->for_share;
+	in_args.object = (char*)get_auth_display_share (actx);
 	in_args.username = (char*)actx->use_user;
 	in_args.domain = (char*)actx->use_domain;
 	in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
@@ -799,10 +823,38 @@ prefill_authentication (SmbAuthContext *actx)
 	DEBUG_SMB(("[auth] Trying to prefill credentials for: %s\n", in_args.uri));
 
 	invoked = gnome_vfs_module_callback_invoke
-		(GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
-		 &in_args, sizeof (in_args),
-		 &out_args, sizeof (out_args));
+		              (GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
+		              &in_args, sizeof (in_args),
+		              &out_args, sizeof (out_args));
 
+        g_free (in_args.uri);
+               
+        /* If that didn't work then try without the share name */
+        if (!invoked || !out_args.valid) {
+
+                g_free (out_args.username);
+                g_free (out_args.domain);
+                g_free (out_args.password);
+      
+                memset (&in_args, 0, sizeof (in_args));
+                in_args.uri = get_auth_display_uri (actx, TRUE);
+                in_args.protocol = "smb";
+                in_args.server = (char*)actx->for_server;
+                in_args.object = (char*)NULL;
+                in_args.username = (char*)actx->use_user;
+                in_args.domain = (char*)actx->use_domain;
+                in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
+
+                memset (&out_args, 0, sizeof (out_args));
+
+                DEBUG_SMB(("[auth] Trying to prefill server credentials for: %s\n", in_args.uri));
+
+                invoked = gnome_vfs_module_callback_invoke
+                                        (GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
+                                        &in_args, sizeof (in_args),
+                                        &out_args, sizeof (out_args));                               
+        }
+        
 	if (invoked && out_args.valid) {
 		g_free (actx->use_user);
 		actx->use_user = string_dup_nzero (out_args.username);
@@ -813,7 +865,6 @@ prefill_authentication (SmbAuthContext *actx)
 		DEBUG_SMB(("[auth] Prefilled credentials: %s@%s:%s\n", actx->use_user, actx->use_domain, actx->use_password));
 	} 
 	
-	g_free (in_args.uri);
 	g_free (out_args.username);
 	g_free (out_args.domain);
 	g_free (out_args.password);
@@ -838,15 +889,15 @@ prompt_authentication (SmbAuthContext *actx)
 	in_args.flags = GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_PASSWORD | GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_SAVING_SUPPORTED;
 	if (actx->state & SMB_AUTH_STATE_PROMPTED)
 		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_PREVIOUS_ATTEMPT_FAILED;
-	if (actx->uri && ((GnomeVFSToplevelURI*)actx->uri)->user_name == NULL) {
+	if (!actx->uri || ((GnomeVFSToplevelURI*)actx->uri)->user_name == NULL) {
 		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME |
 	  			 GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN;
 	}
 
-	in_args.uri = get_auth_display_uri (actx);
+	in_args.uri = get_auth_display_uri (actx, FALSE);
 	in_args.protocol = "smb";
 	in_args.server = (char*)actx->for_server;
-	in_args.object = (char*)actx->for_share;
+	in_args.object = (char*)get_auth_display_share (actx);
 	in_args.username = (char*)actx->use_user;
 	in_args.domain = (char*)actx->use_domain;
 	in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
@@ -864,10 +915,14 @@ prompt_authentication (SmbAuthContext *actx)
 
 	if (invoked) {
 		cancelled = out_args.abort_auth;
-		g_free (actx->use_user);
-		actx->use_user = string_dup_nzero (out_args.username);
-		g_free (actx->use_domain);
-		actx->use_domain = string_dup_nzero (out_args.domain);
+                if (in_args.flags & GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME) {
+                     g_free (actx->use_user);
+                       actx->use_user = string_dup_nzero (out_args.username);
+                }
+                if (in_args.flags & GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN) {
+                      g_free (actx->use_domain);
+                     actx->use_domain = string_dup_nzero (out_args.domain);
+                }
 		g_free (actx->use_password);
 		actx->use_password = out_args.password ? g_strdup (out_args.password) : NULL;
 		g_free (actx->keyring);
@@ -877,7 +932,7 @@ prompt_authentication (SmbAuthContext *actx)
 	} 
 	
 	actx->state |= SMB_AUTH_STATE_PROMPTED;
-	
+
 	g_free (out_args.username);
 	g_free (out_args.domain);
 	g_free (out_args.password);
@@ -894,13 +949,23 @@ save_authentication (SmbAuthContext *actx)
 	GnomeVFSModuleCallbackSaveAuthenticationIn in_args;
 	GnomeVFSModuleCallbackSaveAuthenticationOut out_args;
 	gboolean invoked;
-	
+
+        /* Add to the user cache both with and without shares */
+        if (actx->for_server) {
+                update_user_cache (actx, TRUE);
+                update_user_cache (actx, FALSE);
+        }
+
+        if (!actx->save_auth)
+                return;
+      
+        /* Save with the domain name */
 	memset (&in_args, 0, sizeof (in_args));
-	in_args.uri = get_auth_display_uri (actx);
+	in_args.uri = get_auth_display_uri (actx, TRUE);
 	in_args.keyring = (char*)actx->keyring;
 	in_args.protocol = "smb";
 	in_args.server = (char*)actx->for_server;
-	in_args.object = (char*)actx->for_share;
+	in_args.object = NULL;
 	in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
 	in_args.authtype = NULL;
 	in_args.username = (char*)actx->use_user;
@@ -908,14 +973,38 @@ save_authentication (SmbAuthContext *actx)
 	in_args.password = (char*)actx->use_password;
 
 	memset (&out_args, 0, sizeof (out_args));
-	DEBUG_SMB(("[auth] Saving credentials: %s@%s:%s\n", actx->use_user, actx->use_domain, actx->use_password));
+	DEBUG_SMB(("[auth] Saving credentials: %s = %s@%s:%s\n", in_args.uri, actx->use_user, actx->use_domain, actx->use_password));
 	DEBUG_SMB(("[auth] Keyring: %s\n", actx->keyring));
 
 	invoked = gnome_vfs_module_callback_invoke
-		(GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
-		 &in_args, sizeof (in_args),
-		 &out_args, sizeof (out_args));
+                                (GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
+		                &in_args, sizeof (in_args),
+		                &out_args, sizeof (out_args));
+        
+        g_free (in_args.uri);
+              
+        /* Save without the domain name */
+        memset (&in_args, 0, sizeof (in_args));
+        in_args.uri = get_auth_display_uri (actx, FALSE);
+        in_args.keyring = (char*)actx->keyring;
+        in_args.protocol = "smb";
+        in_args.server = (char*)actx->for_server;
+        in_args.object = (char*)get_auth_display_share (actx);
+        in_args.port = actx->uri ? ((GnomeVFSToplevelURI*)actx->uri)->host_port : 0;
+        in_args.authtype = NULL;
+        in_args.username = (char*)actx->use_user;
+        in_args.domain = (char*)actx->use_domain;
+        in_args.password = (char*)actx->use_password;
+        
+        memset (&out_args, 0, sizeof (out_args));
+        DEBUG_SMB(("[auth] Saving credentials: %s = %s@%s:%s\n", in_args.uri, actx->use_user, actx->use_domain, actx->use_password));
+        DEBUG_SMB(("[auth] Keyring: %s\n", actx->keyring));
 
+        invoked = gnome_vfs_module_callback_invoke
+                                (GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
+                                &in_args, sizeof (in_args),
+                                &out_args, sizeof (out_args));
+        
 	g_free (in_args.uri);
 }
 	
@@ -984,7 +1073,6 @@ init_authentication (SmbAuthContext *actx, GnomeVFSURI *uri)
 	DEBUG_SMB(("[auth] Initializing Authentication\n"));
 	memset (actx, 0, sizeof(*actx));
 	actx->uri = uri;
-	cache_access_failed = FALSE;
 }
 
 static int
@@ -1015,8 +1103,6 @@ perform_authentication (SmbAuthContext *actx)
 		return -1;
 	}
 	
-	/* TODO: What about cache_access_failed? */
-	
 	actx->passes++;
 
 	/* First pass */
@@ -1041,10 +1127,16 @@ perform_authentication (SmbAuthContext *actx)
 		if (!auth_failed) {
 			
 			DEBUG_SMB(("[auth] Operation successful\n"));
-			if (actx->save_auth)
-				save_authentication (actx);
+			save_authentication (actx);
 			ret = 0;
-	
+
+		/* If authentication failed, but we already have a connection ... */
+		} else if (actx->cache_used && !actx->cache_added) {
+
+			/* ... we just return the error */
+			DEBUG_SMB(("[auth] Not reauthenticating a cached connection.\n"));
+			ret = -1;
+
 		/* A failed authentication */
 		} else if (actx->auth_called) {
 			
@@ -1074,8 +1166,8 @@ perform_authentication (SmbAuthContext *actx)
 			if (cont)
 				ret = 1;
 			else {
+				/* Note that we leave actx->res set to whatever it was set to */
 				DEBUG_SMB(("[auth] Authentication cancelled by user.\n"));
-				actx->res = GNOME_VFS_ERROR_CANCELLED;
 				ret = -1;
 			}
 					
@@ -1117,27 +1209,32 @@ auth_callback (const char *server_name, const char *share_name,
 	g_free (actx->for_server);
 	actx->for_server = string_dup_nzero (server_name);
 	g_free (actx->for_share);
-	actx->for_share = !share_name || strcmp (share_name,"IPC$") == 0 
-				? NULL : string_dup_nzero (share_name);
+	actx->for_share = string_dup_nzero (share_name);
 
 	/* The first pass, try the cache, fill in anything we know */
 	if (actx->passes == 1)
 		initial_authentication (actx);
 	
-	/* If we have a valid user and password then go for it */
-	if (actx->use_user && actx->use_password) {
-		DEBUG_SMB(("[auth] Using credentials: %s@%s:%s\n", actx->use_user, actx->use_domain, actx->use_password));
+	/* If we have a valid user then go for it */
+	if (actx->use_user) {
 		strncpy (username_out, actx->use_user, unmaxlen);
-		strncpy (password_out, actx->use_password, pwmaxlen);
-		if (actx->use_domain)
-			strncpy (domain_out, actx->use_domain, domainmaxlen);
+                strncpy (password_out, actx->use_password ? actx->use_password : "", pwmaxlen);
+                strncpy (domain_out, actx->use_domain ? actx->use_domain : "", domainmaxlen);
+                DEBUG_SMB(("[auth] Using credentials: %s:%s@%s\n", username_out, password_out, domain_out));
+
+        /* On first login try a guest login */
+        } else if (actx->passes == 1) {
+                strncpy (username_out, "guest", unmaxlen);
+                strncpy (password_out, "", pwmaxlen);
+                strncpy (domain_out, "", domainmaxlen);
+                DEBUG_SMB(("[auth] No credentials, trying 'guest' user login\n"));
 
 	/* We have no credentials ... */			
 	} else {
-		DEBUG_SMB(("[auth] No credentials, returning null values\n"));
 		strncpy (username_out, "", unmaxlen);
 		strncpy (password_out, "", pwmaxlen);
 		strncpy (domain_out, "", domainmaxlen);
+                DEBUG_SMB(("[auth] No credentials, returning null values\n"));
 	}
 }
 
@@ -2465,7 +2562,7 @@ vfs_module_shutdown (GnomeVFSMethod *method)
 
 	g_hash_table_destroy (server_cache);
 	g_hash_table_destroy (workgroups);
-	g_hash_table_destroy (default_user_hashtable);
+        g_hash_table_destroy (user_cache);
 	
 	g_mutex_free (smb_lock);
 
